#include <iostream>
#include <sstream>
#include <fstream>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cdf.h>
#include <float.h>

#include <math.h>

#include "alleleEst.h"

using namespace std;

void usage(char * name){
  cerr << "\nUSAGE: " << name << " give command line options" << endl;;
  exit(1);
}

/*--------- Functions to determine data dimensions and read in data -------------*/

/* determine and record the number of individuals per locus, based on readsfile */
int getNind(string filename){
  string line;
  int lociCtr = -1;
  int indCtr = 0;
  ifstream infile;

  // open file
  infile.open(filename.c_str());
  if (!infile){
    cerr << "Cannot open file " << filename << endl;
    exit(1);
  }

  while (getline(infile, line)){ // read a line of input into line
    if (MAR == line[0]){ // MAR is the character designating a locus
      if (lociCtr > -1){
	break;
      }
      indCtr = 0;
      lociCtr++;
    }
    else { // this is a line with count data
      indCtr++;
    }
  }
  infile.close();
  return(indCtr);
}

/* determine the number of loci from reads file */
int getNloci(string filename){
  string line;
  int lociCtr = 0;
  ifstream infile;

  // open file
  infile.open(filename.c_str());
  if (!infile){
    cerr << "Cannot open file " << filename << endl;
    exit(1);
  }
  // get number of loci
  while (getline(infile, line)){ // read a line of input into line
    if ( MAR == line[0]){ // MAR is the character designating a locus
      lociCtr++;
    }
  }
  infile.close();
  return(lociCtr);
}

/* determine and record the number of alleles per locus */
int getNallele(string filename, gsl_vector_int * nallele){
  string line, oneword;
  int lociCtr = -1;
  int wordCtr = 0;
  int max = 0;
  int first = 1;
  ifstream infile;

  // open file
  infile.open(filename.c_str());
  if (!infile){
    cerr << "Cannot open file " << filename << endl;
    exit(1);
  }

  // determine number of alleles per locus
  while (getline(infile, line)){ // read a line of input into line
    if (MAR == line[0]){ // MAR is the character designating a locus
      lociCtr++;
      first = 1;
    }
    else if (first == 1){ // this is a line with count data
      first = 0;
      wordCtr = 0;
      istringstream stream(line);
      while (stream >> oneword){ // read a word at a time
	  wordCtr++;
      }
      gsl_vector_int_set(nallele, lociCtr, wordCtr);
      if (wordCtr > max){
	max = wordCtr;
      }
    }
  }
  infile.close();
  return(max);
}

/* Get sequence read data */
void getreads(string readfile, datacont * data){
  int i = -1;
  int n = 0, k = 0;
  int allelecount;
  string line, oneword;
  ifstream infile;


  infile.open(readfile.c_str());
  if (!infile){
    cerr << "Cannot open file " << readfile << endl;
    exit(1);
  }

  // read in data 
  while (getline(infile, line)){ // read a line of input into line
    if (MAR == line[0]){ // MAR is the character designating a locus
      i++; // increment locus
      n = 0;
    }
    else { // this is a line with count data
      k = 0;
      allelecount = 0;
      istringstream stream(line);
      while (stream >> oneword){ // read a word at a time
	gsl_matrix_int_set(data->reads[i].mat, n, k, atoi(oneword.c_str())); // count data 
	if(gsl_matrix_int_get(data->reads[i].mat, n, k) > 0){
	  allelecount++;
	}
	k++;
      } 
      gsl_matrix_int_set(data->nonzeroalleles, i, n, allelecount);
      n++;
    }
  }
  infile.close();
}

/*---------- Functions for parameter init. and MCMC ---------------*/

/* initialize model parameters */
void initparams(datadimcont * datadim, paramcont * params, initcont * initvals, datacont * data){
  int i, j, k, nk;
  int nobs = 0;
  int observedalleles[2];

  // initialize population allele frequencies
  for(i=0; i<datadim->nloci; i++){
    nk=gsl_vector_int_get(datadim->nallele, i);
    gsl_ran_dirichlet(r, nk, params->alleleparams->data, params->alleleresult->data);
    gsl_matrix_set_row(params->popallelefreq, i, params->alleleresult);
  }

  // initialize the genotype frequencies
  for(i=0; i<datadim->nloci; i++){
    nk=gsl_vector_int_get(datadim->nallele, i);
    for(j=0; j<datadim->nind; j++){
      nobs = 0;
      for(k=0; k<nk; k++){
	if(gsl_matrix_int_get(data->reads[i].mat, j, k) > 0){ 
	  // must have this allele, increment by 1
	  observedalleles[nobs] = k;  // store observed allele
	  nobs++;
	}
      }
      // if only a single allele was observed, randomly increment an allele
      k = 0;
      while(nobs < 2){
	observedalleles[nobs] = gsl_rng_uniform_int(r, nk);
	nobs++;
      }
      gsl_matrix_int_set(params->genetics[i][j].mat, observedalleles[0],
			 observedalleles[1], 1);
    }
  }
}

/* mcmc wrapper */
void mcmcUpdate(datadimcont * datadim, paramcont * params, datacont * data){

  updatePopallelefreq(datadim, params); // Gibbs sampling
  updateGenotypes(datadim, params, data); // Gibbs sampling
}

/* gibbs update of pop. allele frequencies */
void updatePopallelefreq(datadimcont * datadim, paramcont * params){
  int i, j, k, l, nk, found;

  for(i=0; i<datadim->nloci; i++){
    nk=gsl_vector_int_get(datadim->nallele, i);
    gsl_vector_set_all(params->alleleparams, 1); // assuming Dirichlet alpha = 1 for all
    for(j=0; j<datadim->nind; j++){
      found = 0;
      for(k=0; k<nk && found<1; k++){
	for(l=k; l<nk && found<1; l++){
	  if(gsl_matrix_int_get(params->genetics[i][j].mat, k, l) == 1){
	    gsl_vector_set(params->alleleparams, k, 
			   gsl_vector_get(params->alleleparams, k)+1); 
	    gsl_vector_set(params->alleleparams, l, 
			   gsl_vector_get(params->alleleparams, l)+1); 
	    found = 1;
	  }
	}
      }
    }
    gsl_ran_dirichlet(r, nk, params->alleleparams->data, params->alleleresult->data);
    gsl_matrix_set_row(params->popallelefreq, i, params->alleleresult);
   }
}

/* gibbs update of genotypes */
void updateGenotypes(datadimcont * datadim, paramcont * params, datacont * data){
  int i, j, k, nk, ctr;
  int sampledalleles[2] = {0, 0};
  double loggenotypeprob = 0;

  for(i=0; i<datadim->nloci; i++){
    nk=gsl_vector_int_get(datadim->nallele, i);
    for(j=0; j<datadim->nind; j++){

      // if gsl_matrix_int_get(data->nonzeroalleles, i, j) == 2
      // we already have the numeric value we need, known heterozygote
      if(gsl_matrix_int_get(data->nonzeroalleles, i, j) == 0){
	gsl_matrix_int_set_zero(params->genetics[i][j].mat);
	gsl_matrix_get_row(params->alleleparams, params->popallelefreq, i);
	gsl_ran_multinomial(r, nk, 2, 
			    params->alleleparams->data, params->allelecount->data);
	for(k=0, ctr=0; k<nk && ctr<2; k++){
	  if(gsl_vector_uint_get(params->allelecount, k) == 2){
	    sampledalleles[0] = k;
	    sampledalleles[1] = k;
	    break;
	  }
	  else if(gsl_vector_uint_get(params->allelecount, k) == 1){
	    sampledalleles[ctr] = k;
	    ctr++;
	  }
	}
	gsl_matrix_int_set(params->genetics[i][j].mat, sampledalleles[0], sampledalleles[1], 1);
      }
      else if(gsl_matrix_int_get(data->nonzeroalleles, i, j) == 1){
	gsl_matrix_int_set_zero(params->genetics[i][j].mat);
	for(k=0; k<nk; k++){
	  if(gsl_matrix_int_get(data->reads[i].mat, j, k) > 0){
	    // gsl_matrix_int_get(data->reads[i].mat, j, k);
	    sampledalleles[0] = k;
	    break;
	  }
	}
	for(k=0; k<nk; k++){
	  if(k == sampledalleles[0]){ //homozygote probability
	    gsl_vector_set(params->alleleparams, k, log(gsl_pow_2(gsl_matrix_get(params->popallelefreq, i, k))));
	  } 
	  else{
	    loggenotypeprob = log(2) + log(gsl_matrix_get(params->popallelefreq, i, k)) +
	      log(gsl_matrix_get(params->popallelefreq, i, sampledalleles[0]));  
	    loggenotypeprob = loggenotypeprob + 
	      log(gsl_ran_binomial_pdf(gsl_matrix_int_get(data->reads[i].mat, j, sampledalleles[0]), 
		  0.5, gsl_matrix_int_get(data->reads[i].mat, j, sampledalleles[0])));
	    // 0.5 above because each k in here involves a heterozygote
	    gsl_vector_set(params->alleleparams, k, loggenotypeprob);
	  }
	}
	for(k=0; k<nk; k++){
	  // exponentiate log probability
	  gsl_vector_set(params->alleleparams, k, exp(gsl_vector_get(params->alleleparams, k)));
	}
	gsl_ran_multinomial(r, nk, 1, 
			    params->alleleparams->data, params->allelecount->data);
	for(k=0; k<nk; k++){
	  if(gsl_vector_uint_get(params->allelecount, k) > 0){
	    sampledalleles[1] = k;
	    break;
	  }
	}
	if(sampledalleles[0] > sampledalleles[1]){  // ensure that we go to upper triangle in matrix
	  k=sampledalleles[1];
	  sampledalleles[1] = sampledalleles[0];
	  sampledalleles[0] = k;
	}
	gsl_matrix_int_set(params->genetics[i][j].mat, sampledalleles[0], sampledalleles[1], 1);
      }
    }
  }
}

/* running sum of genotypes for each individual */
void calcpGenotype(datadimcont * datadim, paramcont * params){
  int i, j, k, l, nk;
  int found;
  
  for(i=0; i<datadim->nloci; i++){
    nk=gsl_vector_int_get(datadim->nallele, i);
    for(j=0; j<datadim->nind; j++){
      found = 0;
      for(k=0; k<nk && found==0; k++){
	for(l=k; l<nk && found==0; l++){
	  if(gsl_matrix_int_get(params->genetics[i][j].mat, k, l) == 1){
	      gsl_matrix_set(params->prgenetics[i][j].mat, k, l, 1 + 
			     gsl_matrix_get(params->prgenetics[i][j].mat, k, l));
	      found = 1;
	  }
	}
      }
    }
  }		     
}

/* ----------- Functions to write results ---------------- */

/* write allele frequency */
void writeSamples(datadimcont * datadim, paramcont * params, FILE * fp){
  int i, k, nk;
  
  for(i=0; i<datadim->nloci; i++){
    nk=gsl_vector_int_get(datadim->nallele, i);
    fprintf(fp, "%i", i);
    for(k=0; k<nk; k++){
      fprintf(fp, " %.5f", gsl_matrix_get(params->popallelefreq, i, k));
    }
    fprintf(fp, "\n");
  }
}

/* writes genotype probs. at the end of the mcmc */
void writeGen(datadimcont * datadim, paramcont * params, FILE * fp, int ctr){
  int i, j, k, l, nk;
  double pg;

  // order is hom 0, het, hom 1
  for(i=0; i<datadim->nloci; i++){
    nk=gsl_vector_int_get(datadim->nallele, i);
    for(j=0; j<datadim->nind; j++){
      fprintf(fp, "%i %i", i, j);
      for(k=0; k<nk; k++){
	for(l=k; l<nk; l++){
	  pg = gsl_matrix_get(params->prgenetics[i][j].mat, k, l) / ctr;
	  fprintf(fp, " %.5f", pg);
	}
      }
      fprintf(fp, "\n");
    }
  }
}
